@@ -3,7 +3,7 @@
  *
  * Copyright (C) 2004, Christophe Saout <christophe@saout.de>
  * Copyright (C) 2004-2007, Clemens Fruhwirth <clemens@endorphin.org>
- * Copyright (C) 2009-2011, Red Hat, Inc. All rights reserved.
+ * Copyright (C) 2009-2012, Red Hat, Inc. All rights reserved.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -16,7 +16,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
 #include <string.h>
@@ -56,6 +56,7 @@ struct crypt_device {
 	char *plain_cipher;
 	char *plain_cipher_mode;
 	char *plain_uuid;
+	unsigned int plain_key_size;
 
 	/* used in CRYPT_LOOPAES */
 	struct crypt_params_loopaes loopaes_hdr;
@@ -357,7 +358,7 @@ static int key_from_terminal(struct crypt_device *cd, char *msg, char **key,
 		} else
 			*key_len = r;
 	} else
-		r = crypt_get_key(msg, key, key_len, 0, NULL, cd->timeout,
+		r = crypt_get_key(msg, key, key_len, 0, 0, NULL, cd->timeout,
 				  (force_verify || cd->password_verify), cd);
 out:
 	free(prompt);
@@ -378,6 +379,9 @@ static int volume_key_by_terminal_passphrase(struct crypt_device *cd, int keyslo
 
 		r = key_from_terminal(cd, NULL, &passphrase_read,
 				      &passphrase_size_read, 0);
+		/* Continue if it is just passphrase verify mismatch */
+		if (r == -EPERM)
+			continue;
 		if(r < 0)
 			goto out;
 
@@ -404,9 +408,10 @@ out:
 
 static int key_from_file(struct crypt_device *cd, char *msg,
 			  char **key, size_t *key_len,
-			  const char *key_file, size_t key_size)
+			  const char *key_file, size_t key_offset,
+			  size_t key_size)
 {
-	return crypt_get_key(msg, key, key_len, key_size, key_file,
+	return crypt_get_key(msg, key, key_len, key_offset, key_size, key_file,
 			     cd->timeout, 0, cd);
 }
 
@@ -597,6 +602,27 @@ int crypt_set_data_device(struct crypt_device *cd, const char *device)
 	return crypt_check_data_device_size(cd);
 }
 
+static int _crypt_load_luks1(struct crypt_device *cd, int require_header, int repair)
+{
+	struct luks_phdr hdr;
+	int r;
+
+	r = init_crypto(cd);
+	if (r < 0)
+		return r;
+
+	r = LUKS_read_phdr(mdata_device(cd), &hdr, require_header, repair, cd);
+	if (r < 0)
+		return r;
+
+	if (!cd->type && !(cd->type = strdup(CRYPT_LUKS1)))
+		return -ENOMEM;
+
+	memcpy(&cd->hdr, &hdr, sizeof(hdr));
+
+	return r;
+}
+
 int crypt_init_by_name_and_header(struct crypt_device **cd,
 				  const char *name,
 				  const char *header_device)
@@ -677,6 +703,7 @@ int crypt_init_by_name_and_header(struct crypt_device **cd,
 		(*cd)->plain_hdr.hash = NULL; /* no way to get this */
 		(*cd)->plain_hdr.offset = dmd.offset;
 		(*cd)->plain_hdr.skip = dmd.iv_offset;
+		(*cd)->plain_key_size = dmd.vk->keylength;
 
 		r = crypt_parse_name_and_mode(dmd.cipher, cipher, NULL, cipher_mode);
 		if (!r) {
@@ -699,7 +726,7 @@ int crypt_init_by_name_and_header(struct crypt_device **cd,
 		}
 	} else if (isLUKS((*cd)->type)) {
 		if (mdata_device(*cd)) {
-			r = crypt_load(*cd, CRYPT_LUKS1, NULL);
+			r = _crypt_load_luks1(*cd, 0, 0);
 			if (r < 0) {
 				log_dbg("LUKS device header does not match active device.");
 				free((*cd)->type);
@@ -754,6 +781,7 @@ static int _crypt_format_plain(struct crypt_device *cd,
 		return -EINVAL;
 	}
 
+	cd->plain_key_size = volume_key_size;
 	cd->volume_key = crypt_alloc_volume_key(volume_key_size, NULL);
 	if (!cd->volume_key)
 		return -ENOMEM;
@@ -928,7 +956,6 @@ int crypt_load(struct crypt_device *cd,
 	       const char *requested_type,
 	       void *params __attribute__((unused)))
 {
-	struct luks_phdr hdr;
 	int r;
 
 	log_dbg("Trying to load %s crypt type from device %s.",
@@ -945,19 +972,41 @@ int crypt_load(struct crypt_device *cd,
 		return -EINVAL;
 	}
 
-	r = init_crypto(cd);
+	r = _crypt_load_luks1(cd, 1, 0);
 	if (r < 0)
 		return r;
 
-	r = LUKS_read_phdr(mdata_device(cd), &hdr, 1, cd);
+	/* cd->type and header must be set in context */
+	r = crypt_check_data_device_size(cd);
+	if (r < 0) {
+		free(cd->type);
+		cd->type = NULL;
+	}
+
+	return r;
+}
+
+int crypt_repair(struct crypt_device *cd,
+		 const char *requested_type,
+		 void *params __attribute__((unused)))
+{
+	int r;
+
+	log_dbg("Trying to repair %s crypt type from device %s.",
+		requested_type ?: "any", mdata_device(cd) ?: "(none)");
+
+	if (!mdata_device(cd))
+		return -EINVAL;
+
+	if (requested_type && !isLUKS(requested_type))
+		return -EINVAL;
+
+
+	/* Load with repair */
+	r = _crypt_load_luks1(cd, 1, 1);
 	if (r < 0)
 		return r;
 
-	if (!cd->type && !(cd->type = strdup(CRYPT_LUKS1)))
-		return -ENOMEM;
-
-	memcpy(&cd->hdr, &hdr, sizeof(hdr));
-
 	/* cd->type and header must be set in context */
 	r = crypt_check_data_device_size(cd);
 	if (r < 0) {
@@ -1197,11 +1246,12 @@ out:
 	return r < 0 ? r : keyslot;
 }
 
-int crypt_resume_by_keyfile(struct crypt_device *cd,
-			    const char *name,
-			    int keyslot,
-			    const char *keyfile,
-			    size_t keyfile_size)
+int crypt_resume_by_keyfile_offset(struct crypt_device *cd,
+				   const char *name,
+				   int keyslot,
+				   const char *keyfile,
+				   size_t keyfile_size,
+				   size_t keyfile_offset)
 {
 	struct volume_key *vk = NULL;
 	char *passphrase_read = NULL;
@@ -1229,7 +1279,8 @@ int crypt_resume_by_keyfile(struct crypt_device *cd,
 		return -EINVAL;
 
 	r = key_from_file(cd, _("Enter passphrase: "), &passphrase_read,
-			  &passphrase_size_read, keyfile, keyfile_size);
+			  &passphrase_size_read, keyfile, keyfile_offset,
+			  keyfile_size);
 	if (r < 0)
 		goto out;
 
@@ -1248,6 +1299,16 @@ out:
 	return r < 0 ? r : keyslot;
 }
 
+int crypt_resume_by_keyfile(struct crypt_device *cd,
+			    const char *name,
+			    int keyslot,
+			    const char *keyfile,
+			    size_t keyfile_size)
+{
+	return crypt_resume_by_keyfile_offset(cd, name, keyslot,
+					      keyfile, keyfile_size, 0);
+}
+
 // slot manipulation
 int crypt_keyslot_add_by_passphrase(struct crypt_device *cd,
 	int keyslot, // -1 any
@@ -1324,12 +1385,14 @@ out:
 	return r ?: keyslot;
 }
 
-int crypt_keyslot_add_by_keyfile(struct crypt_device *cd,
+int crypt_keyslot_add_by_keyfile_offset(struct crypt_device *cd,
 	int keyslot,
 	const char *keyfile,
 	size_t keyfile_size,
+	size_t keyfile_offset,
 	const char *new_keyfile,
-	size_t new_keyfile_size)
+	size_t new_keyfile_size,
+	size_t new_keyfile_offset)
 {
 	struct volume_key *vk = NULL;
 	char *password = NULL; size_t passwordLen;
@@ -1362,7 +1425,7 @@ int crypt_keyslot_add_by_keyfile(struct crypt_device *cd,
 		if (keyfile)
 			r = key_from_file(cd, _("Enter any passphrase: "),
 					  &password, &passwordLen,
-					  keyfile, keyfile_size);
+					  keyfile, keyfile_offset, keyfile_size);
 		else
 			r = key_from_terminal(cd, _("Enter any passphrase: "),
 					      &password, &passwordLen, 0);
@@ -1379,7 +1442,7 @@ int crypt_keyslot_add_by_keyfile(struct crypt_device *cd,
 	if (new_keyfile)
 		r = key_from_file(cd, _("Enter new passphrase for key slot: "),
 				  &new_password, &new_passwordLen, new_keyfile,
-				  new_keyfile_size);
+				  new_keyfile_offset, new_keyfile_size);
 	else
 		r = key_from_terminal(cd, _("Enter new passphrase for key slot: "),
 				      &new_password, &new_passwordLen, 1);
@@ -1395,6 +1458,18 @@ out:
 	return r < 0 ? r : keyslot;
 }
 
+int crypt_keyslot_add_by_keyfile(struct crypt_device *cd,
+	int keyslot,
+	const char *keyfile,
+	size_t keyfile_size,
+	const char *new_keyfile,
+	size_t new_keyfile_size)
+{
+	return crypt_keyslot_add_by_keyfile_offset(cd, keyslot,
+				keyfile, keyfile_size, 0,
+				new_keyfile, new_keyfile_size, 0);
+}
+
 int crypt_keyslot_add_by_volume_key(struct crypt_device *cd,
 	int keyslot,
 	const char *volume_key,
@@ -1516,7 +1591,7 @@ int crypt_activate_by_passphrase(struct crypt_device *cd,
 		}
 
 		r = process_key(cd, cd->plain_hdr.hash,
-				cd->volume_key->keylength,
+				cd->plain_key_size,
 				passphrase, passphrase_size, &vk);
 		if (r < 0)
 			goto out;
@@ -1545,11 +1620,12 @@ out:
 	return r < 0  ? r : keyslot;
 }
 
-int crypt_activate_by_keyfile(struct crypt_device *cd,
+int crypt_activate_by_keyfile_offset(struct crypt_device *cd,
 	const char *name,
 	int keyslot,
 	const char *keyfile,
 	size_t keyfile_size,
+	size_t keyfile_offset,
 	uint32_t flags)
 {
 	crypt_status_info ci;
@@ -1581,12 +1657,12 @@ int crypt_activate_by_keyfile(struct crypt_device *cd,
 
 		r = key_from_file(cd, _("Enter passphrase: "),
 				  &passphrase_read, &passphrase_size_read,
-				  keyfile, keyfile_size);
+				  keyfile, keyfile_offset, keyfile_size);
 		if (r < 0)
 			goto out;
 
 		r = process_key(cd, cd->plain_hdr.hash,
-				cd->volume_key->keylength,
+				cd->plain_key_size,
 				passphrase_read, passphrase_size_read, &vk);
 		if (r < 0)
 			goto out;
@@ -1594,7 +1670,7 @@ int crypt_activate_by_keyfile(struct crypt_device *cd,
 		r = PLAIN_activate(cd, name, vk, cd->plain_hdr.size, flags);
 	} else if (isLUKS(cd->type)) {
 		r = key_from_file(cd, _("Enter passphrase: "), &passphrase_read,
-			  &passphrase_size_read, keyfile, keyfile_size);
+			  &passphrase_size_read, keyfile, keyfile_offset, keyfile_size);
 		if (r < 0)
 			goto out;
 		r = LUKS_open_key_with_hdr(mdata_device(cd), keyslot, passphrase_read,
@@ -1611,7 +1687,7 @@ int crypt_activate_by_keyfile(struct crypt_device *cd,
 		r = keyslot;
 	} else if (isLOOPAES(cd->type)) {
 		r = key_from_file(cd, NULL, &passphrase_read, &passphrase_size_read,
-				  keyfile, keyfile_size);
+				  keyfile, keyfile_offset, keyfile_size);
 		if (r < 0)
 			goto out;
 		r = LOOPAES_parse_keyfile(cd, &vk, cd->loopaes_hdr.hash, &key_count,
@@ -1631,6 +1707,17 @@ out:
 	return r;
 }
 
+int crypt_activate_by_keyfile(struct crypt_device *cd,
+	const char *name,
+	int keyslot,
+	const char *keyfile,
+	size_t keyfile_size,
+	uint32_t flags)
+{
+	return crypt_activate_by_keyfile_offset(cd, name, keyslot, keyfile,
+						keyfile_size, 0, flags);
+}
+
 int crypt_activate_by_volume_key(struct crypt_device *cd,
 	const char *name,
 	const char *volume_key,
@@ -1658,8 +1745,7 @@ int crypt_activate_by_volume_key(struct crypt_device *cd,
 		if (!name)
 			return -EINVAL;
 
-		if (!volume_key || !volume_key_size || !cd->volume_key ||
-			volume_key_size != cd->volume_key->keylength) {
+		if (!volume_key || !volume_key_size || volume_key_size != cd->plain_key_size) {
 			log_err(cd, _("Incorrect volume key specified for plain device.\n"));
 			return -EINVAL;
 		}
@@ -1809,11 +1895,15 @@ void crypt_set_password_retry(struct crypt_device *cd, int tries)
 	cd->tries = tries;
 }
 
-void crypt_set_iterarion_time(struct crypt_device *cd, uint64_t iteration_time_ms)
+void crypt_set_iteration_time(struct crypt_device *cd, uint64_t iteration_time_ms)
 {
 	log_dbg("Iteration time set to %" PRIu64 " miliseconds.", iteration_time_ms);
 	cd->iteration_time = iteration_time_ms;
 }
+void crypt_set_iterarion_time(struct crypt_device *cd, uint64_t iteration_time_ms)
+{
+	crypt_set_iteration_time(cd, iteration_time_ms);
+}
 
 void crypt_set_password_verify(struct crypt_device *cd, int password_verify)
 {
@@ -1976,8 +2066,8 @@ const char *crypt_get_device_name(struct crypt_device *cd)
 
 int crypt_get_volume_key_size(struct crypt_device *cd)
 {
-	if (isPLAIN(cd->type) && cd->volume_key)
-		return cd->volume_key->keylength;
+	if (isPLAIN(cd->type))
+		return cd->plain_key_size;
 
 	if (isLUKS(cd->type))
 		return cd->hdr.keyBytes;
